@@ -48,6 +48,7 @@
 #include "utils/timestamp.h"
 
 #include <string.h>
+#include <stdlib.h>
 
 #include "oracle_fdw.h"
 
@@ -81,6 +82,9 @@ struct OracleFdwOption
 #define OPT_SCHEMA "schema"
 #define OPT_TABLE "table"
 #define OPT_PLAN_COSTS "plan_costs"
+#define OPT_MAX_LONG "max_long"
+
+#define DEFAULT_MAX_LONG 32767
 
 /*
  * Valid options for oracle_fdw.
@@ -92,11 +96,15 @@ static struct OracleFdwOption valid_options[] = {
 	{OPT_PASSWORD, UserMappingRelationId, true},
 	{OPT_SCHEMA, ForeignTableRelationId, false},
 	{OPT_TABLE, ForeignTableRelationId, true},
-	{OPT_PLAN_COSTS, ForeignTableRelationId, false}
+	{OPT_PLAN_COSTS, ForeignTableRelationId, false},
+	{OPT_MAX_LONG, ForeignTableRelationId, false}
 };
 
 #define option_count (sizeof(valid_options)/sizeof(struct OracleFdwOption))
 
+/*
+ * Ways to handle WHERE clauses.
+ */
 typedef enum {
 	FILTER_LOCALLY,
 	PUSHDOWN,
@@ -125,9 +133,9 @@ struct OracleFdwState {
 /*
  * SQL functions
  */
-extern Datum oracle_fdw_handler(PG_FUNCTION_ARGS);
-extern Datum oracle_fdw_validator(PG_FUNCTION_ARGS);
-extern Datum oracle_close_connections(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum oracle_fdw_handler(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum oracle_fdw_validator(PG_FUNCTION_ARGS);
+extern PGDLLEXPORT Datum oracle_close_connections(PG_FUNCTION_ARGS);
 
 PG_FUNCTION_INFO_V1(oracle_fdw_handler);
 PG_FUNCTION_INFO_V1(oracle_fdw_validator);
@@ -136,7 +144,7 @@ PG_FUNCTION_INFO_V1(oracle_close_connections);
 /*
  * on-load initializer
  */
-extern void _PG_init(void);
+extern PGDLLEXPORT void _PG_init(void);
 
 /*
  * FDW callback routines
@@ -167,7 +175,7 @@ static int acquireSampleRowsFunc (Relation relation, int elevel, HeapTuple *rows
 #endif
 static bool getOracleWhereClause(oracleSession *session, char **where, Expr *expr, const struct oraTable *oraTable, struct paramDesc **paramList);
 static void getUsedColumns(Expr *expr, struct oraTable *oraTable);
-static void checkDataTypes(oracleSession *session, struct oraTable *oraTable);
+static void checkDataTypes(struct oraTable *oraTable);
 static char *guessNlsLang(char *nls_lang);
 static List *serializePlanData(struct OracleFdwState *fdwState);
 static Const *serializeString(const char *s);
@@ -185,7 +193,7 @@ static void errorContextCallback(void *arg);
  * Foreign-data wrapper handler function: return a struct with pointers
  * to callback routines.
  */
-Datum
+PGDLLEXPORT Datum
 oracle_fdw_handler(PG_FUNCTION_ARGS)
 {
 	FdwRoutine *fdwroutine = makeNode(FdwRoutine);
@@ -215,7 +223,7 @@ oracle_fdw_handler(PG_FUNCTION_ARGS)
  * 		Raise an ERROR if the option or its value are considered invalid
  * 		or a required option is missing.
  */
-Datum
+PGDLLEXPORT Datum
 oracle_fdw_validator(PG_FUNCTION_ARGS)
 {
 	List *options_list = untransformRelOptions(PG_GETARG_DATUM(0));
@@ -278,6 +286,19 @@ oracle_fdw_validator(PG_FUNCTION_ARGS)
 						errmsg("invalid value for option \"%s\"", def->defname),
 						errhint("Valid values in this context are: on/yes/true or off/no/false")));
 		}
+
+		/* check valid values for max_long */
+		if (strcmp(def->defname, OPT_MAX_LONG) == 0)
+		{
+			char *val = ((Value *) (def->arg))->val.str;
+			char *endptr;
+			unsigned long max_long = strtoul(val, &endptr, 0);
+			if (val[0] == '\0' || *endptr != '\0' || max_long < 1 || max_long > 1073741823ul)
+				ereport(ERROR,
+						(errcode(ERRCODE_FDW_INVALID_ATTRIBUTE_VALUE),
+						errmsg("invalid value for option \"%s\"", def->defname),
+						errhint("Valid values in this context are integers between 1 and 1073741823")));
+		}
 	}
 
 	/* check that all required options have been given */
@@ -298,7 +319,7 @@ oracle_fdw_validator(PG_FUNCTION_ARGS)
  * oracle_close_connections
  * 		Close all open Oracle connections.
  */
-Datum
+PGDLLEXPORT Datum
 oracle_close_connections(PG_FUNCTION_ARGS)
 {
 	elog(DEBUG1, "oracle_fdw: close all Oracle connections");
@@ -313,7 +334,7 @@ oracle_close_connections(PG_FUNCTION_ARGS)
  * 		backend shutdown.
  */
 void
-_PG_init()
+_PG_init(void)
 {
 	on_proc_exit(&exitHook, PointerGetDatum(NULL));
 }
@@ -345,7 +366,7 @@ oraclePlanForeignScan(Oid foreigntableid,
 	elog(DEBUG1, "oracle_fdw: remote query is: %s", fdwState->query);
 
 	/* get PostgreSQL column data types, check that they match Oracle's */
-	checkDataTypes(fdwState->session, fdwState->oraTable);
+	checkDataTypes(fdwState->oraTable);
 
 	/* get Oracle's (bad) estimate only if plan_costs is set */
 	if (plan_costs)
@@ -394,7 +415,7 @@ oracleGetForeignRelSize(PlannerInfo *root, RelOptInfo *baserel, Oid foreigntable
 	elog(DEBUG1, "oracle_fdw: remote query is: %s", fdwState->query);
 
 	/* get PostgreSQL column data types, check that they match Oracle's */
-	checkDataTypes(fdwState->session, fdwState->oraTable);
+	checkDataTypes(fdwState->oraTable);
 
 	/* get Oracle's (bad) estimate only if plan_costs is set */
 	if (plan_costs)
@@ -568,7 +589,7 @@ oracleIterateForeignScan(ForeignScanState *node)
 		elog(DEBUG3, "oracle_fdw: get next row in foreign table scan on %d", RelationGetRelid(node->ss.ss_currentRelation));
 
 		/* fetch the next result row */
-		have_result = oracleFetchNext(fdw_state->session, fdw_state->oraTable);
+		have_result = oracleFetchNext(fdw_state->session);
 	}
 	else
 	{
@@ -650,7 +671,8 @@ struct OracleFdwState
 	char *pgtablename = get_rel_name(foreigntableid);
 	List *options;
 	ListCell *cell;
-	char *schema = NULL, *table = NULL, *plancosts = NULL, *qualtable;
+	char *schema = NULL, *table = NULL, *plancosts = NULL, *maxlong = NULL, *qualtable;
+	long max_long;
 	StringInfoData buf;
 
 	fdwState->nls_lang = NULL;
@@ -682,8 +704,16 @@ struct OracleFdwState
 			table = ((Value *) (def->arg))->val.str;
 		if (strcmp(def->defname, OPT_PLAN_COSTS) == 0)
 			plancosts = ((Value *) (def->arg))->val.str;
+		if (strcmp(def->defname, OPT_MAX_LONG) == 0)
+			maxlong = ((Value *) (def->arg))->val.str;
 	}
 
+	/* convert "max_long" option to number or use default */
+	if (maxlong == NULL)
+		max_long = DEFAULT_MAX_LONG;
+	else
+		max_long = strtol(maxlong, NULL, 0);
+
 	/* check if options are ok */
 	if (table == NULL)
 		ereport(ERROR,
@@ -709,7 +739,7 @@ struct OracleFdwState
 	}
 
 	/* get remote table description */
-	fdwState->oraTable = oracleDescribe(fdwState->session, qualtable, pgtablename);
+	fdwState->oraTable = oracleDescribe(fdwState->session, qualtable, pgtablename, max_long);
 
 	/* add PostgreSQL data to table description */
 	getColumnData(foreigntableid, fdwState->oraTable);
@@ -962,7 +992,7 @@ acquireSampleRowsFunc(Relation relation, int elevel, HeapTuple *rows, int targro
 {
 	int collected_rows = 0, i;
 	struct OracleFdwState *fdw_state;
-	bool plan_costs, first_column = true;;
+	bool plan_costs, first_column = true;
 	StringInfoData query;
 	TupleDesc tupDesc = RelationGetDescr(relation);
 	Datum *values = (Datum *)palloc(tupDesc->natts * sizeof(Datum));
@@ -989,21 +1019,31 @@ acquireSampleRowsFunc(Relation relation, int elevel, HeapTuple *rows, int targro
 	/* loop columns */
 	for (i=0; i<fdw_state->oraTable->ncols; ++i)
 	{
-		/* all columns are used */
-		fdw_state->oraTable->cols[i]->used = 1;
-
-		/* allocate memory for return value */
-		fdw_state->oraTable->cols[i]->val = (char *)palloc(fdw_state->oraTable->cols[i]->val_size);
-		fdw_state->oraTable->cols[i]->val_len = 0;
-		fdw_state->oraTable->cols[i]->val_null = 1;
-
-		if (first_column)
-			first_column = false;
+		/* don't get LONG, LONG RAW and untranslatable values */
+		if (fdw_state->oraTable->cols[i]->oratype == ORA_TYPE_LONG
+				|| fdw_state->oraTable->cols[i]->oratype == ORA_TYPE_LONGRAW
+				|| fdw_state->oraTable->cols[i]->oratype == ORA_TYPE_OTHER)
+		{
+			fdw_state->oraTable->cols[i]->used = 0;
+		}
 		else
-			appendStringInfo(&query, ", ");
-
-		/* append column name */
-		appendStringInfo(&query, "%s", fdw_state->oraTable->cols[i]->name);
+		{
+			/* all columns are used */
+			fdw_state->oraTable->cols[i]->used = 1;
+	
+			/* allocate memory for return value */
+			fdw_state->oraTable->cols[i]->val = (char *)palloc(fdw_state->oraTable->cols[i]->val_size);
+			fdw_state->oraTable->cols[i]->val_len = 0;
+			fdw_state->oraTable->cols[i]->val_null = 1;
+	
+			if (first_column)
+				first_column = false;
+			else
+				appendStringInfo(&query, ", ");
+	
+			/* append column name */
+			appendStringInfo(&query, "%s", fdw_state->oraTable->cols[i]->name);
+		}
 	}
 
 	/* if there are no columns, use NULL */
@@ -1015,11 +1055,11 @@ acquireSampleRowsFunc(Relation relation, int elevel, HeapTuple *rows, int targro
 	elog(DEBUG1, "oracle_fdw: remote query is %s", fdw_state->query);
 
 	/* get PostgreSQL column data types, check that they match Oracle's */
-	checkDataTypes(fdw_state->session, fdw_state->oraTable);
+	checkDataTypes(fdw_state->oraTable);
 
 	/* loop through query results */
 	while(oracleIsStatementOpen(fdw_state->session)
-			? oracleFetchNext(fdw_state->session, fdw_state->oraTable)
+			? oracleFetchNext(fdw_state->session)
 			: oracleExecuteQuery(fdw_state->session, fdw_state->query, fdw_state->oraTable, fdw_state->paramList))
 	{
 		/* allow user to interrupt ANALYZE */
@@ -2153,7 +2193,7 @@ getUsedColumns(Expr *expr, struct oraTable *oraTable)
  * 		converted to the PostgreSQL data types, raise an error if not.
  */
 void
-checkDataTypes(oracleSession *session, struct oraTable *oraTable)
+checkDataTypes(struct oraTable *oraTable)
 {
 	int i;
 
@@ -2169,7 +2209,8 @@ checkDataTypes(oracleSession *session, struct oraTable *oraTable)
 		/* the binary Oracle types can be converted to bytea */
 		if ((oratype == ORA_TYPE_RAW
 				|| oratype == ORA_TYPE_BLOB
-				|| oratype == ORA_TYPE_BFILE)
+				|| oratype == ORA_TYPE_BFILE
+				|| oratype == ORA_TYPE_LONGRAW)
 				&& pgtype == BYTEAOID)
 			continue;
 
@@ -2531,7 +2572,7 @@ struct OracleFdwState
 		state->oraTable->cols[i]->val_size = deserializeLong(lfirst(cell));
 		cell = lnext(cell);
 		/* allocate memory for the result value */
-		state->oraTable->cols[i]->val = (char *)palloc(state->oraTable->cols[i]->val_size);
+		state->oraTable->cols[i]->val = (char *)palloc(state->oraTable->cols[i]->val_size + 1);
 		state->oraTable->cols[i]->val_len = 0;
 		state->oraTable->cols[i]->val_null = 1;
 	}
@@ -2777,11 +2818,21 @@ convertTuple(struct OracleFdwState *fdw_state, Datum *values, bool *nulls, bool
 				|| fdw_state->oraTable->cols[index]->oratype == ORA_TYPE_BFILE
 				|| fdw_state->oraTable->cols[index]->oratype == ORA_TYPE_CLOB)
 		{
-			/* get the actual LOB contents (palloc'ed), truncated if desired */
-			oracleGetLob(fdw_state->session, fdw_state->oraTable,
+			/* for LOBs, get the actual LOB contents (palloc'ed), truncated if desired */
+			oracleGetLob(fdw_state->session,
 				(void *)fdw_state->oraTable->cols[index]->val, fdw_state->oraTable->cols[index]->oratype,
 				&value, &value_len, trunc_lob ? (WIDTH_THRESHOLD+1) : 0);
 		}
+		else if (fdw_state->oraTable->cols[index]->oratype == ORA_TYPE_LONG
+				|| fdw_state->oraTable->cols[index]->oratype == ORA_TYPE_LONGRAW)
+		{
+			/* for LONG and LONG RAW, the first 4 bytes contain the length */
+			value_len = *((long *)fdw_state->oraTable->cols[index]->val);
+			/* the rest is the actual data */
+			value = fdw_state->oraTable->cols[index]->val + 4;
+			/* terminating zero byte (needed for LONGs) */
+			value[value_len] = '\0';
+		}
 		else
 		{
 			/* for other data types, oraTable contains the results */
