@@ -37,6 +37,8 @@ typedef enum
 	ORA_TYPE_BLOB,
 	ORA_TYPE_CLOB,
 	ORA_TYPE_BFILE,
+	ORA_TYPE_LONG,
+	ORA_TYPE_LONGRAW,
 	ORA_TYPE_OTHER
 } oraType;
 
@@ -109,12 +111,12 @@ extern void oracleCloseStatement(oracleSession *session);
 extern void oracleCloseConnections(void);
 extern void oracleShutdown(void);
 extern int oracleIsStatementOpen(oracleSession *session);
-extern struct oraTable *oracleDescribe(oracleSession *session, char *tablename, char *pgname);
+extern struct oraTable *oracleDescribe(oracleSession *session, char *tablename, char *pgname, long max_long);
 extern void oracleEstimate(oracleSession *session, const char *query, double seq_page_cost, int block_size, double *startup_cost, double *total_cost, double *rows, int *width);
 extern void oracleExplain(oracleSession *session, const char *query, int *nrows, char ***plan);
 extern int oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTable *oraTable, struct paramDesc *paramList);
-extern int oracleFetchNext(oracleSession *session, struct oraTable *oraTable);
-extern void oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, oraType type, char **value, long *value_len, unsigned long trunc);
+extern int oracleFetchNext(oracleSession *session);
+extern void oracleGetLob(oracleSession *session, void *locptr, oraType type, char **value, long *value_len, unsigned long trunc);
 extern void oracleCleanupTransaction(void *arg);
 
 /*
