@@ -11,6 +11,10 @@
 #include <sys/types.h>
 #include <unistd.h>
 #include <stdio.h>
+#if defined _WIN32 || defined _WIN64
+/* for getpid */
+#include <process.h>
+#endif
 
 /* Oracle header */
 #include <oci.h>
@@ -30,7 +34,10 @@ static int silent = 0;
 /* contains Oracle error messages, set by checkerr() */
 #define ERRBUFSIZE 1000
 static char oraMessage[ERRBUFSIZE];
-static sb4 errcode;
+static sb4 err_code;
+
+/* set to "1" as soon as OCIEnvCreate is called */
+static int oci_initialized = 0;
 
 /*
  * Linked list for temporary Oracle handles and descriptors.
@@ -97,7 +104,7 @@ struct oracleSession
 static void setOracleEnvironment(char *nls_lang);
 static void oracleQueryPlan(oracleSession *session, const char *query, const char *desc_query, int nres, dvoid **res, sb4 *res_size, ub2 *res_type, ub2 *res_len, sb2 *res_ind);
 static sword checkerr(sword status, dvoid *handle, ub4 handleType);
-static char *copyOraText(const OraText *string, ub4 size, int quote);
+static char *copyOraText(const OraText *string, int size, int quote);
 static void closeSession(OCIEnv *envhp, OCIServer *srvhp, OCISession *userhp, int disconnect);
 static void disconnectServer(OCIEnv *envhp, OCIServer *srvhp);
 static void removeEnvironment(OCIEnv *envhp);
@@ -176,6 +183,9 @@ oracleSession
 				oraMessage);
 		}
 
+		/* we can call OCITerminate now */
+		oci_initialized = 1;
+
 		/* allocate error handle */
 		if (checkerr(
 			OCIHandleAlloc((dvoid *) envhp, (dvoid **) &errhp,
@@ -371,7 +381,7 @@ oracleSession
 		}
 
 		/* set MODULE for the Oracle session */
-		snprintf(pid, 29, "%lu", (unsigned long)getpid());
+		sprintf(pid, "%lu", (unsigned long)getpid());
 		pid[29] = '\0';
 
 		if (checkerr(
@@ -557,7 +567,7 @@ oracleCloseStatement(oracleSession *session)
  * 		Close everything in the cache.
  */
 void
-oracleCloseConnections()
+oracleCloseConnections(void)
 {
 	while (envlist != NULL)
 	{
@@ -579,7 +589,7 @@ oracleCloseConnections()
  * 		This will be called at the end of the PostgreSQL session.
  */
 void
-oracleShutdown()
+oracleShutdown(void)
 {
 	/* don't report error messages */
 	silent = 1;
@@ -587,7 +597,8 @@ oracleShutdown()
 	oracleCloseConnections();
 
 	/* done with Oracle */
-	(void)OCITerminate(OCI_DEFAULT);
+	if (oci_initialized)
+		(void)OCITerminate(OCI_DEFAULT);
 }
 
 /*
@@ -606,7 +617,7 @@ oracleIsStatementOpen(oracleSession *session)
  * 		Returns a palloc'ed data structure with the results.
  */
 struct oraTable
-*oracleDescribe(oracleSession *session, char *tablename, char *pgname)
+*oracleDescribe(oracleSession *session, char *tablename, char *pgname, long max_long)
 {
 	struct oraTable *reply;
 	char *mytablename = tablename;
@@ -651,7 +662,7 @@ struct oraTable
 				(ub4)strlen(mytablename), OCI_OTYPE_NAME, OCI_DEFAULT, OCI_PTYPE_UNK, dschp),
 			(dvoid *)session->envp->errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
 		{
-			if (errcode == 4043)
+			if (err_code == 4043)
 				oracleError_sd(FDW_TABLE_NOT_FOUND,
 					"remote table for \"%s\" does not exist or does not allow read access", pgname,
 					oraMessage);
@@ -697,7 +708,7 @@ struct oraTable
 					oraMessage);
 			}
 
-			name = copyOraText(ident, ident_size, 1);
+			name = copyOraText(ident, (int)ident_size, 1);
 
 			/* referenced table schema */
 			if (checkerr(
@@ -710,7 +721,7 @@ struct oraTable
 					oraMessage);
 			}
 
-			schema = copyOraText(ident, ident_size, 1);
+			schema = copyOraText(ident, (int)ident_size, 1);
 
 			/* referenced table database link */
 			if (checkerr(
@@ -723,7 +734,7 @@ struct oraTable
 					oraMessage);
 			}
 
-			link = copyOraText(ident, ident_size, 0);
+			link = copyOraText(ident, (int)ident_size, 0);
 
 			/* construct a table name for the next lookup */
 			synonym_len = strlen(name);
@@ -769,7 +780,7 @@ struct oraTable
 				(ub4)strlen(tablename), OCI_OTYPE_NAME, OCI_DEFAULT, OCI_PTYPE_TABLE, dschp),
 			(dvoid *)session->envp->errhp, OCI_HTYPE_ERROR) != OCI_SUCCESS)
 		{
-			if (errcode == 4043)
+			if (err_code == 4043)
 				oracleError_sd(FDW_TABLE_NOT_FOUND,
 					"remote table for \"%s\" does not exist or does not allow read access", pgname,
 					oraMessage);
@@ -850,7 +861,7 @@ struct oraTable
 				oraMessage);
 		}
 
-		reply->cols[i-1]->name = copyOraText(ident, ident_size, 1);
+		reply->cols[i-1]->name = copyOraText(ident, (int)ident_size, 1);
 
 		/* get the data type */
 		if (checkerr(
@@ -1015,7 +1026,16 @@ struct oraTable
 				reply->cols[i-1]->oratype = ORA_TYPE_INTERVALD2S;
 				reply->cols[i-1]->val_size = precision + scale + 12;
 				break;
-			case SQLT_LVB:
+			case SQLT_LBI:
+				/* LONG RAW */
+				reply->cols[i-1]->oratype = ORA_TYPE_LONGRAW;
+				reply->cols[i-1]->val_size = max_long + 4;
+				break;
+			case SQLT_LNG:
+				/* LONG */
+				reply->cols[i-1]->oratype = ORA_TYPE_LONG;
+				reply->cols[i-1]->val_size = max_long + 4;
+				break;
 			case SQLT_BIN:
 				/* RAW(n) */
 				reply->cols[i-1]->oratype = ORA_TYPE_RAW;
@@ -1371,7 +1391,7 @@ oracleQueryPlan(oracleSession *session, const char *query, const char *desc_quer
 	{
 
 		/* create a better message if we lack permissions on V$SQL */
-		if (errcode == 942)
+		if (err_code == 942)
 			oracleError_d(FDW_UNABLE_TO_CREATE_EXECUTION,
 				"no SELECT privilege on V$SQL in the remote database",
 				oraMessage);
@@ -1450,7 +1470,7 @@ oracleQueryPlan(oracleSession *session, const char *query, const char *desc_quer
 	{
 
 		/* create a better message if we lack permissions on V$SQL_PLAN */
-		if (errcode == 942)
+		if (err_code == 942)
 			oracleError_d(FDW_UNABLE_TO_CREATE_EXECUTION,
 				"no SELECT privilege on V$SQL_PLAN in the remote database",
 				oraMessage);
@@ -1645,6 +1665,12 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
 					else
 						type = SQLT_BIN;
 					break;
+				case ORA_TYPE_LONG:
+					type = SQLT_LVC;
+					break;
+				case ORA_TYPE_LONGRAW:
+					type = SQLT_LVB;
+					break;
 				default:
 					/* all other columns are converted to strings */
 					type = SQLT_STR;
@@ -1714,7 +1740,7 @@ oracleExecuteQuery(oracleSession *session, const char *query, const struct oraTa
  * 		Fetch the next result row, return 1 if there is one, else 0.
  */
 int
-oracleFetchNext(oracleSession *session, struct oraTable *oraTable)
+oracleFetchNext(oracleSession *session)
 {
 	sword result;
 
@@ -1742,11 +1768,10 @@ oracleFetchNext(oracleSession *session, struct oraTable *oraTable)
 /*
  * oracleGetLob
  * 		Get the LOB contents and store them in *value and *value_len.
- * 		"oraTable" is passed so that all LOB locators can be freed in case of errors.
  * 		If "trunc" is nonzero, it contains the number of bytes or characters to get.
  */
 void
-oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, oraType type, char **value, long *value_len, unsigned long trunc)
+oracleGetLob(oracleSession *session, void *locptr, oraType type, char **value, long *value_len, unsigned long trunc)
 {
 	OCILobLocator *locp = *(OCILobLocator **)locptr;
 	oraub8 amount_byte, amount_char;
@@ -1801,7 +1826,7 @@ oracleGetLob(oracleSession *session, struct oraTable *oraTable, void *locptr, or
 		}
 
 		/* update LOB length */
-		*value_len += amount_byte;
+		*value_len += (long)amount_byte;
 	}
 	while (result == OCI_NEED_DATA);
 
@@ -1866,7 +1891,7 @@ checkerr(sword status, dvoid *handle, ub4 handleType)
 	oraMessage[0] = '\0';
 
 	if (status == OCI_SUCCESS_WITH_INFO || status == OCI_ERROR)
-		OCIErrorGet(handle, (ub4)1, NULL, &errcode,
+		OCIErrorGet(handle, (ub4)1, NULL, &err_code,
 			(text *)oraMessage, (ub4)ERRBUFSIZE, handleType);
 
 	if (status == OCI_SUCCESS_WITH_INFO)
@@ -1875,7 +1900,7 @@ checkerr(sword status, dvoid *handle, ub4 handleType)
 	if (status == OCI_NO_DATA)
 	{
 		strcpy(oraMessage, "ORA-00100: no data found");
-		errcode = (sb4)100;
+		err_code = (sb4)100;
 	}
 
 	return status;
@@ -1886,7 +1911,7 @@ checkerr(sword status, dvoid *handle, ub4 handleType)
  * 		Returns a palloc'ed string containing a (possibly quoted) copy of "string".
  */
 char
-*copyOraText(const OraText *string, ub4 size, int quote)
+*copyOraText(const OraText *string, int size, int quote)
 {
 	int resultsize = (quote ? size + 2 : size);
 	register int i, j=-1;
